from abc import ABC, abstractmethod
import copy
import os
from time import time

import pandas as pd
import numpy as np
import yaml
from tqdm import tqdm

import torch
import torch.nn as nn
from torch.utils.data import Dataset, DataLoader
from torch.optim.lr_scheduler import ReduceLROnPlateau

from tensorboardX import SummaryWriter

from netam.common import (
    generate_kmers,
    kmer_to_index_of,
    nt_mask_tensor_of,
    optimizer_of_name,
    tensor_to_np_if_needed,
    BASES,
    BASES_AND_N_TO_INDEX,
    BIG,
    VRC01_NT_SEQ,
    encode_sequences,
)
from netam import models
import netam.molevol as molevol


def encode_mut_pos_and_base(parent, child, site_count=None):
    """
    This function takes a parent and child sequence and returns a tuple of
    tensors: (mutation_indicator, new_base_idxs).
    The mutation_indicator tensor is a boolean tensor indicating whether
    each site is mutated. Both the parent and the child must be one of
    A, C, G, T, to be considered a mutation.
    The new_base_idxs tensor is an integer tensor that gives the index of the
    new base at each site.

    Note that we use -1 as a placeholder for non-mutated bases in the
    new_base_idxs tensor. This ensures that lack of masking will lead
    to an error.

    If site_count is not None, then the tensors will be trimmed & padded to the
    provided length.
    """
    assert len(parent) == len(child), f"{parent} and {child} are not the same length"

    if site_count is None:
        site_count = len(parent)

    mutation_indicator = []
    new_base_idxs = []

    for i in range(min(len(parent), site_count)):
        if parent[i] != child[i] and parent[i] in BASES and child[i] in BASES:
            mutation_indicator.append(1)
            new_base_idxs.append(BASES_AND_N_TO_INDEX[child[i]])
        else:
            mutation_indicator.append(0)
            new_base_idxs.append(-1)  # No mutation, so set to -1

    # Pad the lists if necessary
    if len(mutation_indicator) < site_count:
        padding_length = site_count - len(mutation_indicator)
        mutation_indicator += [0] * padding_length
        new_base_idxs += [-1] * padding_length

    return (
        torch.tensor(mutation_indicator, dtype=torch.bool),
        torch.tensor(new_base_idxs, dtype=torch.int64),
    )


def wt_base_modifier_of(parent, site_count):
    """The wt_base_modifier tensor is all 0s except for the wt base at each site, which
    is -BIG.

    We will add wt_base_modifier to the CSP logits. This will zero out the prediction of
    WT at each site after softmax.
    """
    wt_base_modifier = torch.zeros((site_count, 4))
    for i, base in enumerate(parent[:site_count]):
        if base in BASES:
            wt_base_modifier[i, BASES_AND_N_TO_INDEX[base]] = -BIG
    return wt_base_modifier


class KmerSequenceEncoder:
    def __init__(self, kmer_length, site_count):
        self.kmer_length = kmer_length
        self.site_count = site_count
        assert kmer_length % 2 == 1
        self.overhang_length = (kmer_length - 1) // 2
        self.all_kmers = generate_kmers(kmer_length)
        self.kmer_to_index = kmer_to_index_of(self.all_kmers)

    @property
    def parameters(self):
        return {"kmer_length": self.kmer_length, "site_count": self.site_count}

    def encode_sequence(self, sequence):
        sequence = sequence.upper()
        # Pad sequence with overhang_length 'N's at the start and end so that we
        # can assign parameters to every site in the sequence.
        padded_sequence = (
            "N" * self.overhang_length + sequence + "N" * self.overhang_length
        )

        # Note that we are using a default value of 0 here. So we use the
        # catch-all term for anything with an N in it for the sites on the
        # boundary of the kmer.
        # Note that this line also effectively pads things out to site_count because
        # when i gets large the slice will be empty and we will get a 0.
        # These sites will get masked out by the mask below.
        kmer_indices = [
            self.kmer_to_index.get(padded_sequence[i : i + self.kmer_length], 0)
            for i in range(self.site_count)
        ]

        wt_base_modifier = wt_base_modifier_of(sequence, self.site_count)

        return torch.tensor(kmer_indices, dtype=torch.int32), wt_base_modifier


class PlaceholderEncoder:
    @property
    def parameters(self):
        return {}


class BranchLengthDataset(Dataset):
    def __len__(self):
        return len(self.branch_lengths)

    def export_branch_lengths(self, out_csv_path):
        pd.DataFrame({"branch_length": self.branch_lengths}).to_csv(
            out_csv_path, index=False
        )

    def load_branch_lengths(self, in_csv_path):
        self.branch_lengths = torch.Tensor(
            pd.read_csv(in_csv_path)["branch_length"].values
        )

    def __repr__(self):
        return f"{self.__class__.__name__}(Size: {len(self)}) on {self.branch_lengths.device}"


class SHMoofDataset(BranchLengthDataset):
    def __init__(self, dataframe, kmer_length, site_count):
        super().__init__()
        self.encoder = KmerSequenceEncoder(kmer_length, site_count)
        (
            self.encoded_parents,
            self.masks,
            self.mutation_indicators,
            self.new_base_idxs,
            self.wt_base_modifier,
            self.branch_lengths,
        ) = self.encode_pcps(dataframe)
        assert self.encoded_parents.shape[0] == self.branch_lengths.shape[0]

    def __getitem__(self, idx):
        return (
            self.encoded_parents[idx],
            self.masks[idx],
            self.mutation_indicators[idx],
            self.new_base_idxs[idx],
            self.wt_base_modifier[idx],
            self.branch_lengths[idx],
        )

    def to(self, device):
        self.encoded_parents = self.encoded_parents.to(device)
        self.masks = self.masks.to(device)
        self.mutation_indicators = self.mutation_indicators.to(device)
        self.new_base_idxs = self.new_base_idxs.to(device)
        self.wt_base_modifier = self.wt_base_modifier.to(device)
        self.branch_lengths = self.branch_lengths.to(device)

    def encode_pcps(self, dataframe):
        encoded_parents = []
        masks = []
        mutation_vectors = []
        new_base_idx_vectors = []
        wt_base_modifier_vectors = []
        branch_lengths = []

        for _, row in dataframe.iterrows():
            encoded_parent, wt_base_modifier = self.encoder.encode_sequence(
                row["parent"]
            )
            mask = nt_mask_tensor_of(row["child"], self.encoder.site_count)
            # Assert that anything that is masked in the child is also masked in
            # the parent. We only use the parent_mask for this check.
            parent_mask = nt_mask_tensor_of(row["parent"], self.encoder.site_count)
            assert (mask <= parent_mask).all()
            (
                mutation_indicator,
                new_base_idxs,
            ) = encode_mut_pos_and_base(
                row["parent"], row["child"], self.encoder.site_count
            )

            encoded_parents.append(encoded_parent)
            masks.append(mask)
            mutation_vectors.append(mutation_indicator)
            new_base_idx_vectors.append(new_base_idxs)
            wt_base_modifier_vectors.append(wt_base_modifier)
            # The initial branch lengths are the normalized number of mutations.
            branch_lengths.append(mutation_indicator.sum() / mask.sum())

        return (
            torch.stack(encoded_parents),
            torch.stack(masks),
            torch.stack(mutation_vectors),
            torch.stack(new_base_idx_vectors),
            torch.stack(wt_base_modifier_vectors),
            torch.tensor(branch_lengths),
        )

    def normalized_mutation_frequency(self):
        return self.mutation_indicators.sum(axis=1) / self.masks.sum(axis=1)

    def export_branch_lengths(self, out_csv_path):
        pd.DataFrame(
            {
                "branch_length": tensor_to_np_if_needed(self.branch_lengths),
                "mut_freq": tensor_to_np_if_needed(
                    self.normalized_mutation_frequency()
                ),
            }
        ).to_csv(out_csv_path, index=False)


class Crepe:
    """A lightweight wrapper around a model that can be used for prediction but not
    training.

    It handles serialization.
    """

    SERIALIZATION_VERSION = 0

    def __init__(self, encoder, model, training_hyperparameters={}):
        super().__init__()
        self.encoder = encoder
        self.model = model
        self.training_hyperparameters = training_hyperparameters

    def __call__(self, sequences):
        """Evaluate the model on a list of sequences."""
        if isinstance(sequences, str):
            raise ValueError(
                "Expected a list of sequences for call on crepe, but got a single string instead."
            )
        return self.model.evaluate_sequences(sequences, encoder=self.encoder)

    @property
    def device(self):
        return next(self.model.parameters()).device

    def to(self, device):
        self.model.to(device)

    def encode_sequences(self, sequences):
        return encode_sequences(sequences, self.encoder)

    def save(self, prefix):
        torch.save(self.model.state_dict(), f"{prefix}.pth")
        with open(f"{prefix}.yml", "w") as f:
            yaml.dump(
                {
                    "serialization_version": self.SERIALIZATION_VERSION,
                    "model_class": self.model.__class__.__name__,
                    "model_hyperparameters": self.model.hyperparameters,
                    "training_hyperparameters": self.training_hyperparameters,
                    "encoder_class": self.encoder.__class__.__name__,
                    "encoder_parameters": self.encoder.parameters,
                },
                f,
            )


def load_crepe(prefix, device=None):
    assert crepe_exists(prefix), f"Crepe {prefix} not found."
    with open(f"{prefix}.yml", "r") as f:
        config = yaml.safe_load(f)

    if config["serialization_version"] != Crepe.SERIALIZATION_VERSION:
        raise ValueError(
            f"Unsupported serialization version: {config['serialization_version']}"
        )

    encoder_class_name = config["encoder_class"]

    try:
        encoder_class = globals()[encoder_class_name]
    except AttributeError:
        raise ValueError(f"Encoder class '{encoder_class_name}' not known.")

    encoder = encoder_class(**config["encoder_parameters"])

    model_class_name = config["model_class"]

    try:
        model_class = getattr(models, model_class_name)
    except AttributeError:
        raise ValueError(
            f"Model class '{model_class_name}' not found in 'models' module."
        )

    model = model_class(**config["model_hyperparameters"])

    model_state_path = f"{prefix}.pth"
    if device is None:
        device = torch.device("cpu")
    model.load_state_dict(
        torch.load(model_state_path, map_location=device, weights_only=True)
    )
    model.eval()

    crepe_instance = Crepe(encoder, model, config["training_hyperparameters"])
    if device is not None:
        crepe_instance.to(device)

    return crepe_instance


def crepe_exists(prefix):
    return os.path.exists(f"{prefix}.yml") and os.path.exists(f"{prefix}.pth")


def trimmed_shm_model_outputs_of_crepe(crepe, parents):
    """Model outputs trimmed to the length of the parent sequences."""
    rates, csp_logits = crepe(parents)
    rates = rates.cpu().detach()
    csps = torch.softmax(csp_logits, dim=-1).cpu().detach()
    trimmed_rates = [rates[i, : len(parent)] for i, parent in enumerate(parents)]
    trimmed_csps = [csps[i, : len(parent)] for i, parent in enumerate(parents)]
    return trimmed_rates, trimmed_csps


def load_pcp_df(pcp_df_path_gz, sample_count=None, chosen_v_families=None):
    """Load a PCP dataframe from a gzipped CSV file.

    `orig_pcp_idx` is the index column from the original file, even if we subset by
    sampling or by choosing V families.
    """
    pcp_df = (
        pd.read_csv(pcp_df_path_gz, compression="gzip", index_col=0)
        .reset_index()
        .rename(columns={"index": "orig_pcp_idx"})
    )
    pcp_df["v_family"] = pcp_df["v_gene"].str.split("-").str[0]
    if chosen_v_families is not None:
        chosen_v_families = set(chosen_v_families)
        pcp_df = pcp_df[pcp_df["v_family"].isin(chosen_v_families)]
    if sample_count is not None:
        pcp_df = pcp_df.sample(sample_count)
    pcp_df.reset_index(drop=True, inplace=True)
    return pcp_df


def add_shm_model_outputs_to_pcp_df(pcp_df, crepe):
    rates, csps = trimmed_shm_model_outputs_of_crepe(crepe, pcp_df["parent"])
    pcp_df["nt_rates"] = rates
    pcp_df["nt_csps"] = csps
    return pcp_df


class Burrito(ABC):
    def __init__(
        self,
        train_dataset,
        val_dataset,
        model,
        optimizer_name="RMSprop",
        batch_size=1024,
        learning_rate=0.1,
        min_learning_rate=1e-4,
        weight_decay=1e-6,
        name="",
    ):
        """Note that we allow train_dataset to be None, to support use cases where we
        just want to do evaluation."""
        self.train_dataset = train_dataset
        self.val_dataset = val_dataset
        if train_dataset is not None:
            self.writer = SummaryWriter(log_dir=f"./_logs/{name}")
            self.writer.add_text("model_name", model.__class__.__name__)
            self.writer.add_text("model_hyperparameters", str(model.hyperparameters))
        self.model = model
        self.optimizer_name = optimizer_name
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.min_learning_rate = min_learning_rate
        self.weight_decay = weight_decay
        self.name = name
        self.reset_optimization()
        self.bce_loss = nn.BCELoss()
        self.global_epoch = 0
        self.start_time = time()

    def build_train_loader(self):
        if self.train_dataset is None:
            return None
        else:
            return DataLoader(
                self.train_dataset, batch_size=self.batch_size, shuffle=True
            )

    def build_val_loader(self):
        return DataLoader(self.val_dataset, batch_size=self.batch_size, shuffle=False)

    @property
    def device(self):
        return next(self.model.parameters()).device

    def reset_optimization(self, learning_rate=None):
        """Reset the optimizer and scheduler."""
        if learning_rate is None:
            learning_rate = self.learning_rate

        # copied from https://github.com/karpathy/nanoGPT/blob/9755682b981a45507f6eb9b11eadef8cb83cebd5/model.py#L264
        param_dict = {
            pn: p for pn, p in self.model.named_parameters() if p.requires_grad
        }
        # Do not apply weight decay to 1D parameters (biases and layernorm weights).
        decay_params = [p for p in param_dict.values() if p.dim() >= 2]
        nodecay_params = [p for p in param_dict.values() if p.dim() < 2]
        optim_groups = [
            {"params": decay_params, "weight_decay": self.weight_decay},
            {"params": nodecay_params, "weight_decay": 0.0},
        ]

        self.optimizer = optimizer_of_name(
            self.optimizer_name,
            optim_groups,
            lr=learning_rate,
            weight_decay=self.weight_decay,
        )
        self.scheduler = ReduceLROnPlateau(
            self.optimizer, mode="min", factor=0.5, patience=10
        )

    def execution_time(self):
        """Return time since the Burrito was created."""
        return time() - self.start_time

    def multi_train(self, epochs, max_tries=3):
        """Train the model.

        If lr isn't below min_lr, reset the optimizer and scheduler, and reset the model
        and resume training.
        """
        for i in range(max_tries):
            train_history = self.simple_train(epochs)
            if self.optimizer.param_groups[0]["lr"] < self.min_learning_rate:
                return train_history
            else:
                print(
                    f"Learning rate {self.optimizer.param_groups[0]['lr']} not below {self.min_learning_rate}. Resetting model and optimizer."
                )
                self.reset_optimization()
                self.model.reinitialize_weights()
        print(f"Failed to train model to min_learning_rate after {max_tries} tries.")
        return train_history

    def write_loss(self, loss_name, loss, step):
        self.writer.add_scalar(loss_name, loss, step, walltime=self.execution_time())

    def write_cuda_memory_info(self):
        megabyte_scaling_factor = 1 / 1024**2
        if self.device.type == "cuda":
            self.writer.add_scalar(
                "CUDA memory allocated",
                torch.cuda.memory_allocated() * megabyte_scaling_factor,
                self.global_epoch,
            )
            self.writer.add_scalar(
                "CUDA max memory allocated",
                torch.cuda.max_memory_allocated() * megabyte_scaling_factor,
                self.global_epoch,
            )

    def process_data_loader(self, data_loader, train_mode=False, loss_reduction=None):
        """Process data through the model using the given data loader. If train_mode is
        True, performs optimization steps.

        Args:
            data_loader (DataLoader): DataLoader to use.
            train_mode (bool, optional): Whether to do optimization as part of
                the forward pass. Defaults to False.
                Note that this also applies the regularization loss if set to True.
            loss_reduction (callable, optional): Function to reduce the loss
                tensor to a scalar. If None, uses torch.sum. Defaults to None.

        Returns:
            float: Average loss.
        """
        total_loss = None
        total_samples = 0

        if train_mode:
            self.model.train()
        else:
            self.model.eval()

        if loss_reduction is None:
            loss_reduction = torch.sum

        with torch.set_grad_enabled(train_mode):
            for batch in data_loader:
                loss = self.loss_of_batch(batch)
                if total_loss is None:
                    total_loss = torch.zeros_like(loss)

                if train_mode:
                    max_grad_retries = 5
                    for grad_retry_count in range(max_grad_retries):
                        scalar_loss = loss_reduction(loss)
                        if hasattr(self.model, "regularization_loss"):
                            reg_loss = self.model.regularization_loss()
                            scalar_loss += reg_loss

                        self.optimizer.zero_grad()
                        scalar_loss.backward()

                        nan_in_gradients = False
                        for name, param in self.model.named_parameters():
                            if torch.isnan(param).any():
                                raise ValueError(f"NaN in weights: {name}")
                            if param.grad is not None and torch.isnan(param.grad).any():
                                nan_in_gradients = True

                        if not nan_in_gradients:
                            self.optimizer.step()
                            break
                        else:
                            if grad_retry_count < max_grad_retries - 1:
                                print(
                                    f"Retrying gradient calculation ({grad_retry_count + 1}/{max_grad_retries}) with loss {torch.sum(loss).item()}"
                                )
                                loss = self.loss_of_batch(batch)
                            else:
                                raise ValueError(f"Exceeded maximum gradient retries!")

                # We support both dicts and lists of tensors as the batch.
                first_value_of_batch = (
                    list(batch.values())[0] if isinstance(batch, dict) else batch[0]
                )
                batch_size = len(first_value_of_batch)
                # If we multiply the loss by the batch size, then the loss will be the sum of the
                # losses for each example in the batch. Then, when we divide by the number of
                # examples in the dataset below, we will get the average loss per example.
                total_loss += loss.detach() * batch_size
                total_samples += batch_size

        average_loss = total_loss / total_samples
        if hasattr(self, "writer"):
            if train_mode:
                self.write_loss("Training loss", average_loss, self.global_epoch)
            else:
                self.write_loss("Validation loss", average_loss, self.global_epoch)
        return loss_reduction(average_loss)

    def simple_train(self, epochs, out_prefix=None):
        """Train the model for the given number of epochs.

        If out_prefix is provided, then a crepe will be saved to that location.
        """
        assert self.train_dataset is not None, "No training data provided."
        train_loader = self.build_train_loader()
        val_loader = self.build_val_loader()

        train_losses = []
        val_losses = []
        best_val_loss = float("inf")
        best_model_state = None

        def record_losses(train_loss, val_loss):
            train_losses.append(train_loss)
            val_losses.append(val_loss)

        with tqdm(range(1, epochs + 1), desc="Epoch") as pbar:
            for epoch in pbar:
                current_lr = self.optimizer.param_groups[0]["lr"]
                if (
                    isinstance(self.scheduler, ReduceLROnPlateau)
                    and current_lr < self.min_learning_rate
                ):
                    break

                if self.device.type == "cuda":
                    # Clear cache for accurate memory usage tracking.
                    torch.cuda.empty_cache()

                train_loss = self.process_data_loader(
                    train_loader, train_mode=True
                ).item()
                val_loss = self.process_data_loader(val_loader, train_mode=False).item()
                self.scheduler.step(val_loss)
                record_losses(train_loss, val_loss)
                self.global_epoch += 1

                if val_loss < best_val_loss:
                    best_val_loss = val_loss
                    best_model_state = copy.deepcopy(self.model.state_dict())

                current_lr = self.optimizer.param_groups[0]["lr"]
                if len(val_losses) > 1:
                    val_loss = val_losses[-1]
                    loss_diff = val_losses[-1] - val_losses[-2]
                    pbar.set_postfix(
                        val_loss=f"{val_loss:.4g}",
                        loss_diff=f"{loss_diff:.4g}",
                        lr=current_lr,
                        refresh=True,
                    )
                self.write_cuda_memory_info()
                self.writer.flush()

        if best_model_state is not None:
            self.model.load_state_dict(best_model_state)

        if out_prefix is not None:
            self.save_crepe(out_prefix)

        return pd.DataFrame({"train_loss": train_losses, "val_loss": val_losses})

    def evaluate(self):
        """Evaluate the model on the validation set."""
        val_loader = self.build_val_loader()
        return self.process_data_loader(val_loader, train_mode=False).item()

    def save_crepe(self, prefix):
        self.to_crepe().save(prefix)

    def standardize_model_rates(self):
        """This is an opportunity to standardize the model rates.

        Only the SHMBurrito class implements this, which makes sense because it needs to
        get normalized but the DNSM does not.
        """
        pass

    def standardize_and_optimize_branch_lengths(self, **optimization_kwargs):
        self.standardize_model_rates()
        if "learning_rate" not in optimization_kwargs:
            optimization_kwargs["learning_rate"] = 0.01
        if "optimization_tol" not in optimization_kwargs:
            optimization_kwargs["optimization_tol"] = 1e-3
        # We do the branch length optimization on CPU but want to restore the
        # model to the device it was on before.
        device = self.device
        self.model.to("cpu")
        for dataset in [self.train_dataset, self.val_dataset]:
            if dataset is None:
                continue
            dataset.to("cpu")
            dataset.branch_lengths = self.find_optimal_branch_lengths(
                dataset, **optimization_kwargs
            )
            dataset.to(device)
        self.model.to(device)

    def standardize_and_use_yun_approx_branch_lengths(self):
        """Yun Song's approximation to the branch lengths.

        This approximation is the mutation count divided by the total mutation rate for
        the sequence. See
        https://github.com/matsengrp/netam/assets/112708/034abb74-5635-48dc-bf28-4321b9110222
        """
        self.standardize_model_rates()
        for dataset in [self.train_dataset, self.val_dataset]:
            if dataset is None:
                continue
            lengths = []
            for (
                encoded_parent,
                mask,
                mutation_indicator,
                wt_base_modifier,
            ) in zip(
                dataset.encoded_parents,
                dataset.masks,
                dataset.mutation_indicators,
                dataset.wt_base_modifier,
            ):
                rates, _ = self.model(
                    encoded_parent.unsqueeze(0),
                    mask.unsqueeze(0),
                    wt_base_modifier.unsqueeze(0),
                )
                mutation_indicator = mutation_indicator[mask].float()
                length = torch.sum(mutation_indicator) / torch.sum(rates)
                lengths.append(length.item())
            dataset.branch_lengths = torch.tensor(lengths)

    def mark_branch_lengths_optimized(self, cycle):
        self.writer.add_scalar(
            "branch length optimization",
            cycle,
            self.global_epoch,
            walltime=self.execution_time(),
        )

    def joint_train(
        self,
        epochs=100,
        cycle_count=2,
        training_method="full",
        out_prefix=None,
        optimize_bl_first_cycle=True,
    ):
        """Do joint optimization of model and branch lengths.

        If training_method is "full", then we optimize the branch lengths using full ML
        optimization. If training_method is "yun", then we use Yun's approximation to
        the branch lengths. If training_method is "fixed", then we fix the branch
        lengths and only optimize the model.

        We give an option to optimize the branch lengths in the first cycle (by default
        we do). But, this can be useful to turn off e.g. if we've loaded in some
        preoptimized branch lengths.

        We reset the optimization after each cycle, and we use a learning rate schedule
        that uses a weighted geometric mean of the current learning rate and the initial
        learning rate that progressively moves towards keeping the current learning rate
        as the cycles progress.
        """
        if training_method == "full":
            optimize_branch_lengths = self.standardize_and_optimize_branch_lengths
        elif training_method == "yun":
            optimize_branch_lengths = self.standardize_and_use_yun_approx_branch_lengths
        elif training_method == "fixed":
            optimize_branch_lengths = lambda: None
        else:
            raise ValueError(f"Unknown training method {training_method}")
        loss_history_l = []
        if optimize_bl_first_cycle:
            optimize_branch_lengths()
        self.mark_branch_lengths_optimized(0)
        for cycle in range(cycle_count):
            print(
                f"### Beginning cycle {cycle + 1}/{cycle_count} using optimizer {self.optimizer_name}"
            )
            current_lr = self.optimizer.param_groups[0]["lr"]
            # set new_lr to be the geometric mean of current_lr and the
            # originally-specified learning rate
            weight = 0.5 + cycle / (2 * cycle_count)
            new_lr = np.exp(
                weight * np.log(current_lr) + (1 - weight) * np.log(self.learning_rate)
            )
            self.reset_optimization(new_lr)
            loss_history_l.append(self.simple_train(epochs, out_prefix=out_prefix))
            # We standardize and optimize the branch lengths after each cycle, even the last one.
            optimize_branch_lengths()
            self.mark_branch_lengths_optimized(cycle + 1)

        return pd.concat(loss_history_l, ignore_index=True)

    @abstractmethod
    def loss_of_batch(self, batch):
        pass

    @abstractmethod
    def to_crepe(self):
        pass


class SHMBurrito(Burrito):
    def __init__(
        self,
        train_dataset,
        val_dataset,
        model,
        optimizer_name="RMSprop",
        batch_size=1024,
        learning_rate=0.1,
        min_learning_rate=1e-4,
        weight_decay=1e-6,
        name="",
    ):
        super().__init__(
            train_dataset,
            val_dataset,
            model,
            optimizer_name=optimizer_name,
            batch_size=batch_size,
            learning_rate=learning_rate,
            min_learning_rate=min_learning_rate,
            weight_decay=weight_decay,
            name=name,
        )

    def loss_of_batch(self, batch):
        (
            encoded_parents,
            masks,
            mutation_indicators,
            _,
            wt_base_modifier,
            branch_lengths,
        ) = batch
        rates = self.model(encoded_parents, masks, wt_base_modifier)
        mut_prob = 1 - torch.exp(-rates * branch_lengths.unsqueeze(-1))
        mut_prob_masked = mut_prob[masks]
        mutation_indicator_masked = mutation_indicators[masks].float()
        loss = self.bce_loss(mut_prob_masked, mutation_indicator_masked)
        return loss

    def vrc01_site_14_model_rate(self):
        """Calculate rate on site 14 (zero-indexed) of VRC01_NT_SEQ."""
        encoder = self.val_dataset.encoder
        assert (
            encoder.site_count >= 15
        ), "Encoder site count too small vrc01_site_14_model_rate"
        encoded_parent, wt_base_modifier = encoder.encode_sequence(VRC01_NT_SEQ)
        mask = nt_mask_tensor_of(VRC01_NT_SEQ, encoder.site_count)
        encoded_parent = encoded_parent.to(self.device)
        mask = mask.to(self.device)
        wt_base_modifier = wt_base_modifier.to(self.device)
        vrc01_rates, _ = self.model(
            encoded_parent.unsqueeze(0),
            mask.unsqueeze(0),
            wt_base_modifier.unsqueeze(0),
        )
        vrc01_rate_14 = vrc01_rates.squeeze()[14].item()
        return vrc01_rate_14

    def standardize_model_rates(self):
        """Normalize the rates output by the model so that it predicts rate 1 on site 14
        (zero-indexed) of VRC01_NT_SEQ."""
        vrc01_rate_14 = self.vrc01_site_14_model_rate()
        self.model.adjust_rate_bias_by(-np.log(vrc01_rate_14))

    def to_crepe(self):
        training_hyperparameters = {
            key: self.__dict__[key]
            for key in [
                "learning_rate",
                "min_learning_rate",
                "weight_decay",
            ]
        }
        encoder = KmerSequenceEncoder(
            self.model.hyperparameters["kmer_length"],
            self.train_dataset.encoder.site_count,
        )
        return Crepe(encoder, self.model, training_hyperparameters)


class TwoLossMixin:
    """A mixin for models that have two losses, one for mutation position and one for
    conditional substitution probability (CSP)."""

    def process_data_loader(self, data_loader, train_mode=False, loss_reduction=None):
        if loss_reduction is None:
            loss_reduction = lambda x: torch.sum(x * self.loss_weights)

        return super().process_data_loader(data_loader, train_mode, loss_reduction)

    def write_loss(self, loss_name, loss, step):
        rate_loss, csp_loss = loss.unbind()
        self.writer.add_scalar(
            "Mut pos " + loss_name,
            rate_loss.item(),
            step,
            walltime=self.execution_time(),
        )
        self.writer.add_scalar(
            "CSP " + loss_name, csp_loss.item(), step, walltime=self.execution_time()
        )


class RSSHMBurrito(TwoLossMixin, SHMBurrito):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.xent_loss = nn.CrossEntropyLoss()
        self.loss_weights = torch.tensor([1.0, 0.01]).to(self.device)

    def evaluate(self):
        val_loader = self.build_val_loader()
        return super().process_data_loader(
            val_loader, train_mode=False, loss_reduction=lambda x: x
        )

    def loss_of_batch(self, batch):
        (
            encoded_parents,
            masks,
            mutation_indicators,
            new_base_idxs,
            wt_base_modifier,
            branch_lengths,
        ) = batch
        rates, csp_logits = self.model(encoded_parents, masks, wt_base_modifier)

        mut_prob = 1 - torch.exp(-rates * branch_lengths.unsqueeze(-1))
        mut_prob_masked = mut_prob[masks]
        mutation_indicator_masked = mutation_indicators[masks].float()
        mut_pos_loss = self.bce_loss(mut_prob_masked, mutation_indicator_masked)

        # Conditional substitution probability (CSP) loss calculation
        # Mask the new_base_idxs to focus only on positions with mutations
        mutated_positions_mask = mutation_indicators == 1
        csp_logits_masked = csp_logits[mutated_positions_mask]
        new_base_idxs_masked = new_base_idxs[mutated_positions_mask]
        # Recall that WT bases are encoded as -1 in new_base_idxs_masked, so
        # this assert makes sure that the loss is masked out for WT bases.
        assert (new_base_idxs_masked >= 0).all()
        csp_loss = self.xent_loss(csp_logits_masked, new_base_idxs_masked)

        return torch.stack([mut_pos_loss, csp_loss])

    def _find_optimal_branch_length(
        self,
        encoded_parent,
        mask,
        mutation_indicator,
        wt_base_modifier,
        starting_branch_length,
        **optimization_kwargs,
    ):
        if torch.sum(mutation_indicator) == 0:
            return 0.0, False

        rates, _ = self.model(
            encoded_parent.unsqueeze(0),
            mask.unsqueeze(0),
            wt_base_modifier.unsqueeze(0),
        )

        rates = rates.squeeze().double()
        mutation_indicator_masked = mutation_indicator[mask].double()

        def log_pcp_probability(log_branch_length):
            branch_length = torch.exp(log_branch_length)
            mut_prob = 1 - torch.exp(-rates * branch_length)
            mut_prob_masked = mut_prob[mask]
            rate_loss = self.bce_loss(mut_prob_masked, mutation_indicator_masked)
            return -rate_loss

        return molevol.optimize_branch_length(
            log_pcp_probability,
            starting_branch_length.double().item(),
            **optimization_kwargs,
        )

    def find_optimal_branch_lengths(self, dataset, **optimization_kwargs):
        optimal_lengths = []
        failed_count = 0

        self.model.eval()
        self.model.freeze()

        for (
            encoded_parent,
            mask,
            mutation_indicator,
            wt_base_modifier,
            starting_branch_length,
        ) in tqdm(
            zip(
                dataset.encoded_parents,
                dataset.masks,
                dataset.mutation_indicators,
                dataset.wt_base_modifier,
                dataset.branch_lengths,
            ),
            total=len(dataset.encoded_parents),
            desc="Finding optimal branch lengths",
        ):
            branch_length, failed_to_converge = self._find_optimal_branch_length(
                encoded_parent,
                mask,
                mutation_indicator,
                wt_base_modifier,
                starting_branch_length,
                **optimization_kwargs,
            )

            optimal_lengths.append(branch_length)
            failed_count += failed_to_converge

        if failed_count > 0:
            print(
                f"Branch length optimization failed to converge for {failed_count} of {len(dataset)} sequences."
            )

        self.model.unfreeze()

        return torch.tensor(optimal_lengths)


def burrito_class_of_model(model):
    if isinstance(model, models.RSCNNModel):
        return RSSHMBurrito
    else:
        return SHMBurrito



"""Free functions for molecular evolution computation.

CSP means conditional substitution probability. CSPs are the probabilities of alternate
states conditioned on there being a substitution.
"""

import numpy as np

import torch
from torch import Tensor, optim

from netam.sequences import CODON_AA_INDICATOR_MATRIX

import netam.sequences as sequences


def check_csps(parent_idxs: Tensor, csps: Tensor) -> Tensor:
    """Make sure that the CSPs are valid, i.e. that they are a probability distribution
    and the parent state is zero.

    Args:
        parent_idxs (torch.Tensor): The parent sequence indices.
        sub_probs (torch.Tensor): A 2D PyTorch tensor representing substitution
            probabilities. Rows correspond to sites, and columns correspond
            to states (e.g. nucleotides).
    """

    # Assert that sub_probs are within the range [0, 1] modulo rounding error
    assert torch.all(csps >= -1e-6), "Substitution probabilities must be non-negative"
    assert torch.all(
        csps <= 1 + 1e-6
    ), "Substitution probabilities must be less than or equal to 1"

    # Create an array of row indices that matches the shape of `parent_idxs`.
    row_indices = torch.arange(len(parent_idxs))

    # Assert that the parent entry has a substitution probability of nearly 0.
    assert torch.all(
        csps[row_indices, parent_idxs] < 1e-6
    ), "Parent entry must have a substitution probability of nearly 0"
    assert torch.allclose(
        csps[: len(parent_idxs)].sum(dim=1), torch.ones(len(parent_idxs))
    )


def build_mutation_matrices(
    parent_codon_idxs: Tensor, codon_mut_probs: Tensor, codon_csps: Tensor
) -> Tensor:
    """Generate a sequence of 3x4 mutation matrices for parent codons along a sequence.

    Given indices for parent codons, mutation probabilities, and substitution
    probabilities for each parent codon along the sequence, this function
    constructs a sequence of 3x4 matrices. Each matrix in the sequence
    represents the mutation probabilities for each nucleotide position in a
    parent codon. The ijkth entry of the resulting tensor corresponds to the
    probability of the jth nucleotide in the ith parent codon mutating to the
    kth nucleotide (in indices).

    Args:
        parent_codon_idxs (torch.Tensor): 2D tensor with each row containing indices representing
            the parent codon's nucleotides at each site along the sequence.
            Shape should be (codon_count, 3).
        codon_mut_probs (torch.Tensor): 2D tensor representing the mutation probabilities for each site in the codon,
            for each codon along the sequence. Shape should be (codon_count, 3).
        codon_csps (torch.Tensor): 3D tensor representing conditional substitution probabilities for each NT site of each codon along the
            sequence. Shape should be (codon_count, 3, 4).

    Returns:
        torch.Tensor: A 4D tensor with shape (codon_count, 3, 4) where the ijkth entry is the mutation probability
            of the jth position in the ith parent codon mutating to the kth nucleotide.
    """

    codon_count = parent_codon_idxs.shape[0]
    assert parent_codon_idxs.shape[1] == 3, "Each parent codon must be of length 3"

    result_matrices = torch.empty((codon_count, 3, 4))

    # Create a mask with the shape (codon_count, 3, 4) to identify where
    # nucleotides in the parent codon are the same as the nucleotide positions
    # in the new codon. Each row in the third dimension contains a boolean
    # value, which is True if the nucleotide position matches the parent codon
    # nucleotide. How it works: None adds one more dimension to the tensor, so
    # that the shape of the tensor is (codon_count, 3, 1) instead of
    # (codon_count, 3). Then broadcasting automatically expands dimensions where
    # needed. So the arange(4) tensor is automatically expanded to match the
    # (codon_count, 3, 1) shape by implicitly turning it into a (1, 1, 4) shape
    # tensor, where it is then broadcasted to the shape (codon_count, 3, 4) to
    # match the shape of parent_codon_idxs[:, :, None] for equality
    # testing.
    mask_same_nt = torch.arange(4) == parent_codon_idxs[:, :, None]

    # Find the multi-dimensional indices where the nucleotide in the parent
    # codon is the same as the nucleotide in the mutation outcome (i.e., no
    # mutation occurs).
    same_nt_indices = torch.nonzero(mask_same_nt)

    # Using the multi-dimensional indices obtained from the boolean mask, update
    # the mutation probability in result_matrices to be "1.0 -
    # mutation_probability" at these specific positions. This captures the
    # probability of a given nucleotide not mutating.
    result_matrices[
        same_nt_indices[:, 0], same_nt_indices[:, 1], same_nt_indices[:, 2]
    ] = (1.0 - codon_mut_probs[same_nt_indices[:, 0], same_nt_indices[:, 1]])

    # Assign values where the nucleotide is different via broadcasting.
    mask_diff_nt = ~mask_same_nt
    result_matrices[mask_diff_nt] = (codon_mut_probs[:, :, None] * codon_csps)[
        mask_diff_nt
    ]

    return result_matrices


def codon_probs_of_mutation_matrices(mut_matrix: Tensor) -> Tensor:
    """Compute the probability tensor for mutating to the codon ijk along the entire
    sequence.

    Args:
    mut_matrix (torch.Tensor): A 3D tensor representing the mutation matrix for the entire sequence.
                               The shape should be (n_sites, 3, 4), where n_sites is the number of sites,
                               3 is the number of positions in a codon, and 4 is the number of nucleotides.

    Returns:
    torch.Tensor: A 4D tensor where the first axis represents different sites in the sequence and
                  the ijk-th entry of the remaining 3D tensor is the probability of mutating to the codon ijk.
    """
    assert (
        mut_matrix.shape[1] == 3
    ), "The second dimension of the input mut_matrix should be 3 to represent the 3 positions in a codon"
    assert (
        mut_matrix.shape[2] == 4
    ), "The last dimension of the input mut_matrix should be 4 to represent the 4 nucleotides"

    # The key to understanding how this works is that when these tensors are
    # multiplied, PyTorch broadcasts them into a common shape (n_sites, 4, 4, 4),
    # performing element-wise multiplication for each slice along the first axis
    # (i.e., for each site).
    return (
        mut_matrix[:, 0, :, None, None]
        * mut_matrix[:, 1, None, :, None]
        * mut_matrix[:, 2, None, None, :]
    )


def aaprobs_of_codon_probs(codon_probs: Tensor) -> Tensor:
    """Compute the probability of each amino acid from the probability of each codon,
    for each parent codon along the sequence.

    Args:
    codon_probs (torch.Tensor): A 4D tensor representing the probability of mutating
                                to each codon for each parent codon along the sequence.
                                Shape should be (codon_count, 4, 4, 4).

    Returns:
    torch.Tensor: A 2D tensor with shape (codon_count, 20) where the ij-th entry is the probability
                  of mutating to the amino acid j from the codon i for each parent codon along the sequence.
    """
    codon_count = codon_probs.shape[0]

    # Reshape such that we merge the last three dimensions into a single dimension while keeping
    # the `codon_count` dimension intact. This prepares the tensor for matrix multiplication.
    reshaped_probs = codon_probs.reshape(codon_count, -1)

    # Perform matrix multiplication to get unnormalized amino acid probabilities.
    aaprobs = torch.matmul(reshaped_probs, CODON_AA_INDICATOR_MATRIX)

    # Normalize probabilities along the amino acid dimension.
    row_sums = aaprobs.sum(dim=1, keepdim=True)
    aaprobs /= row_sums

    return aaprobs


def aaprob_of_mut_and_sub(
    parent_codon_idxs: Tensor, codon_mut_probs: Tensor, codon_csps: Tensor
) -> Tensor:
    """For a sequence of parent codons and given nucleotide mutability and substitution
    probabilities, compute the probability of a substitution to each amino acid for each
    codon along the sequence.

    This function actually isn't used anymore, but there is a good test for it, which
    tests other functions, so we keep it.

    Stop codons don't appear as part of this calculation.

    Args:
    parent_codon_idxs (torch.Tensor): A 2D tensor where each row contains indices representing
                                      the parent codon's nucleotides at each site along the sequence.
                                      Shape should be (codon_count, 3).
    codon_mut_probs (torch.Tensor): A 2D tensor representing the mutation probabilities for each site in the codon,
                              for each codon along the sequence. Shape should be (codon_count, 3).
    codon_csps (torch.Tensor): A 3D tensor representing conditional substitution probabilities for each NT site of each codon along the
                                sequence.
                                Shape should be (codon_count, 3, 4).

    Returns:
    torch.Tensor: A 2D tensor with shape (codon_count, 20) where the ij-th entry is the probability
                  of mutating to the amino acid j from the codon i for each parent codon along the sequence.
    """
    mut_matrices = build_mutation_matrices(
        parent_codon_idxs, codon_mut_probs, codon_csps
    )
    codon_probs = codon_probs_of_mutation_matrices(mut_matrices)
    return aaprobs_of_codon_probs(codon_probs)


def reshape_for_codons(array: Tensor) -> Tensor:
    """Reshape a tensor to add a codon dimension by taking groups of 3 sites.

    Args:
    array (torch.Tensor): Original tensor.

    Returns:
    torch.Tensor: Reshaped tensor with an added codon dimension.
    """
    site_count = array.shape[0]
    assert site_count % 3 == 0, "Site count must be a multiple of 3"
    codon_count = site_count // 3
    return array.reshape(codon_count, 3, *array.shape[1:])


def codon_probs_of_parent_scaled_nt_rates_and_csps(
    parent_idxs: torch.Tensor, scaled_nt_rates: torch.Tensor, nt_csps: torch.Tensor
):
    """Compute the probabilities of mutating to various codons for a parent sequence.

    This uses the same machinery as we use for fitting the DNSM, but we stay on
    the codon level rather than moving to syn/nonsyn changes.

    Args:
        parent_idxs (torch.Tensor): The parent nucleotide sequence encoded as a
            tensor of length Cx3, where C is the number of codons, containing the nt indices of each site.
        scaled_nt_rates (torch.Tensor): Poisson rates of mutation per site, scaled by branch length.
        nt_csps (torch.Tensor): Conditional substitution probabilities per site: a 2D tensor with shape (site_count, 4).

    Returns:
        torch.Tensor: A 4D tensor with shape (codon_count, 4, 4, 4) where the cijk-th entry is the probability
            of the c'th codon mutating to the codon ijk.
    """
    mut_probs = 1.0 - torch.exp(-scaled_nt_rates)
    parent_codon_idxs = reshape_for_codons(parent_idxs)
    codon_mut_probs = reshape_for_codons(mut_probs)
    codon_csps = reshape_for_codons(nt_csps)

    mut_matrices = build_mutation_matrices(
        parent_codon_idxs, codon_mut_probs, codon_csps
    )
    codon_probs = codon_probs_of_mutation_matrices(mut_matrices)

    return codon_probs


def aaprobs_of_parent_scaled_rates_and_csps(
    parent_idxs: Tensor, scaled_nt_rates: Tensor, nt_csps: Tensor
) -> Tensor:
    """Calculate per-site amino acid probabilities from per-site nucleotide rates and
    substitution probabilities.

    Args:
        parent_idxs (torch.Tensor): Parent nucleotide indices. Shape should be (site_count,).
        scaled_nt_rates (torch.Tensor): Poisson rates of mutation per site, scaled by branch length.
                                     Shape should be (site_count,).
        nt_csps (torch.Tensor): Substitution probabilities per site: a 2D
                                  tensor with shape (site_count, 4).

    Returns:
        torch.Tensor: A 2D tensor with rows corresponding to sites and columns
                      corresponding to amino acids.
    """
    return aaprobs_of_codon_probs(
        codon_probs_of_parent_scaled_nt_rates_and_csps(
            parent_idxs, scaled_nt_rates, nt_csps
        )
    )


def build_codon_mutsel(
    parent_codon_idxs: Tensor,
    codon_mut_probs: Tensor,
    codon_csps: Tensor,
    aa_sel_matrices: Tensor,
    multihit_model=None,
) -> Tensor:
    """Build a sequence of codon mutation-selection matrices for codons along a
    sequence.

    These will assign zero for the probability of mutating to a stop codon.

    Args:
        parent_codon_idxs (torch.Tensor): The parent codons for each sequence. Shape: (codon_count, 3)
        codon_mut_probs (torch.Tensor): The mutation probabilities for each site in each codon. Shape: (codon_count, 3)
        codon_csps (torch.Tensor): The substitution probabilities for each site in each codon. Shape: (codon_count, 3, 4)
        aa_sel_matrices (torch.Tensor): The amino-acid selection matrices for each site. Shape: (codon_count, 20)

    Returns:
        torch.Tensor: The probability of mutating to each codon, for each sequence. Shape: (codon_count, 4, 4, 4)
    """
    mut_matrices = build_mutation_matrices(
        parent_codon_idxs, codon_mut_probs, codon_csps
    )
    codon_probs = codon_probs_of_mutation_matrices(mut_matrices)

    if multihit_model is not None:
        codon_probs = multihit_model(parent_codon_idxs, codon_probs)

    # Calculate the codon selection matrix for each sequence via Einstein
    # summation, in which we sum over the repeated indices.
    # So, for each site (s) and codon (c), sum over amino acids (a):
    # codon_sel_matrices[s, c] = sum_a(CODON_AA_INDICATOR_MATRIX[c, a] * aa_sel_matrices[s, a])
    # Resulting shape is (S, C) where S is the number of sites and C is the number of codons.
    # Stop codons don't appear in this sum, so columns for stop codons will be zero.
    codon_sel_matrices = torch.einsum(
        "ca,sa->sc", CODON_AA_INDICATOR_MATRIX, aa_sel_matrices
    )

    # Multiply the codon probabilities by the selection matrices
    codon_mutsel = codon_probs * codon_sel_matrices.view(-1, 4, 4, 4)

    # Clamp the codon_mutsel above by 1: these are probabilities.
    codon_mutsel = codon_mutsel.clamp(max=1.0)

    # Now we need to recalculate the probability of staying in the same codon.
    # In our setup, this is the probability of nothing happening.
    # To calculate this, we zero out the probabilities of mutating to the parent
    # codon...
    codon_count = parent_codon_idxs.shape[0]
    codon_mutsel[(torch.arange(codon_count), *parent_codon_idxs.T)] = 0.0
    # sum together their probabilities...
    sums = codon_mutsel.sum(dim=(1, 2, 3))
    # then set the parent codon probabilities to 1 minus the sum.
    codon_mutsel[(torch.arange(codon_count), *parent_codon_idxs.T)] = 1.0 - sums
    codon_mutsel = codon_mutsel.clamp(min=0.0)

    if sums.max() > 1.0:
        sums_too_big = sums.max()
    else:
        sums_too_big = None

    return codon_mutsel, sums_too_big


def neutral_aa_probs(
    parent_codon_idxs: Tensor,
    codon_mut_probs: Tensor,
    codon_csps: Tensor,
    multihit_model=None,
) -> Tensor:
    """For every site, what is the probability that the amino acid will mutate to every
    amino acid?

    Args:
        parent_codon_idxs (torch.Tensor): The parent codons for each sequence. Shape: (codon_count, 3)
        codon_mut_probs (torch.Tensor): The mutation probabilities for each site in each codon. Shape: (codon_count, 3)
        codon_csps (torch.Tensor): The substitution probabilities for each site in each codon. Shape: (codon_count, 3, 4)

    Returns:
        torch.Tensor: The probability that each site will change to each amino acid.
                      Shape: (codon_count, 20)
    """

    mut_matrices = build_mutation_matrices(
        parent_codon_idxs, codon_mut_probs, codon_csps
    )
    codon_probs = codon_probs_of_mutation_matrices(mut_matrices)

    if multihit_model is not None:
        codon_probs = multihit_model(parent_codon_idxs, codon_probs)

    # Get the probability of mutating to each amino acid.
    aa_probs = codon_probs.view(-1, 64) @ CODON_AA_INDICATOR_MATRIX

    return aa_probs


def mut_probs_of_aa_probs(
    parent_codon_idxs: Tensor,
    aa_probs: Tensor,
) -> Tensor:
    """For every site, what is the probability that the amino acid will have a
    substution or mutate to a stop under neutral evolution?

    Args:
        parent_codon_idxs (torch.Tensor): The parent codons for each sequence. Shape: (codon_count, 3)
        aa_probs (torch.Tensor): The probability that each site will change to each amino acid. Shape: (codon_count, 20)
    """
    # We build a table that will allow us to look up the amino acid index
    # from the codon indices. Argmax gets the aa index.
    aa_idx_from_codon = CODON_AA_INDICATOR_MATRIX.argmax(dim=1).view(4, 4, 4)

    # Get the amino acid index for each parent codon.
    parent_aa_idxs = aa_idx_from_codon[
        (
            parent_codon_idxs[:, 0],
            parent_codon_idxs[:, 1],
            parent_codon_idxs[:, 2],
        )
    ]
    p_staying_same = aa_probs[(torch.arange(len(parent_aa_idxs)), parent_aa_idxs)]

    return 1.0 - p_staying_same


def neutral_aa_mut_probs(
    parent_codon_idxs: Tensor,
    codon_mut_probs: Tensor,
    codon_csps: Tensor,
    multihit_model=None,
) -> Tensor:
    """For every site, what is the probability that the amino acid will have a
    substution or mutate to a stop under neutral evolution?

    This code computes all the probabilities and then indexes into that tensor
    to get the relevant probabilities. This isn't the most efficient way to do
    this, but it's the cleanest. We could make it faster as needed.

    Args:
        parent_codon_idxs (torch.Tensor): The parent codons for each sequence. Shape: (codon_count, 3)
        codon_mut_probs (torch.Tensor): The mutation probabilities for each site in each codon. Shape: (codon_count, 3)
        codon_csps (torch.Tensor): The conditional substitution probabilities for each site in each codon. Shape: (codon_count, 3, 4)

    Returns:
        torch.Tensor: The probability that each site will change to some other amino acid.
                      Shape: (codon_count,)
    """

    aa_probs = neutral_aa_probs(
        parent_codon_idxs,
        codon_mut_probs,
        codon_csps,
        multihit_model=multihit_model,
    )
    mut_probs = mut_probs_of_aa_probs(parent_codon_idxs, aa_probs)
    return mut_probs


def mutsel_log_pcp_probability_of(
    sel_matrix, parent, child, nt_rates, nt_csps, multihit_model=None
):
    """Constructs the log_pcp_probability function specific to given nt_rates and
    nt_csps.

    This function takes log_branch_length as input and returns the log probability of
    the child sequence. It uses log of branch length to ensure non-negativity.
    """

    assert len(parent) % 3 == 0
    assert sel_matrix.shape == (len(parent) // 3, 20)

    parent_idxs = sequences.nt_idx_tensor_of_str(parent)
    child_idxs = sequences.nt_idx_tensor_of_str(child)

    def log_pcp_probability(log_branch_length: torch.Tensor):
        branch_length = torch.exp(log_branch_length)
        nt_mut_probs = 1.0 - torch.exp(-branch_length * nt_rates)

        codon_mutsel, sums_too_big = build_codon_mutsel(
            parent_idxs.reshape(-1, 3),
            nt_mut_probs.reshape(-1, 3),
            nt_csps.reshape(-1, 3, 4),
            sel_matrix,
            multihit_model=multihit_model,
        )

        # This is a diagnostic generating data for netam issue #7.
        # if sums_too_big is not None:
        #     self.csv_file.write(f"{parent},{child},{branch_length},{sums_too_big}\n")

        reshaped_child_idxs = child_idxs.reshape(-1, 3)
        child_prob_vector = codon_mutsel[
            torch.arange(len(reshaped_child_idxs)),
            reshaped_child_idxs[:, 0],
            reshaped_child_idxs[:, 1],
            reshaped_child_idxs[:, 2],
        ]

        child_prob_vector = torch.clamp(child_prob_vector, min=1e-10)

        result = torch.sum(torch.log(child_prob_vector))

        assert torch.isfinite(result)

        return result

    return log_pcp_probability


def optimize_branch_length(
    log_prob_fn,
    starting_branch_length,
    learning_rate=0.1,
    max_optimization_steps=1000,
    optimization_tol=1e-3,
    log_branch_length_lower_threshold=-10.0,
):
    log_branch_length = torch.tensor(np.log(starting_branch_length), requires_grad=True)

    optimizer = optim.Adam([log_branch_length], lr=learning_rate)
    prev_log_branch_length = log_branch_length.clone()

    step_idx = 0

    for step_idx in range(max_optimization_steps):
        # For some PCPs, the optimizer works very hard optimizing very tiny branch lengths.
        if log_branch_length < log_branch_length_lower_threshold:
            break

        optimizer.zero_grad()

        loss = -log_prob_fn(log_branch_length)
        assert torch.isfinite(
            loss
        ), f"Loss is not finite on step {step_idx}: perhaps selection has given a probability of zero?"
        loss.backward()
        torch.nn.utils.clip_grad_norm_([log_branch_length], max_norm=5.0)
        optimizer.step()
        if torch.isnan(log_branch_length):
            raise ValueError("branch length optimization resulted in NAN")

        change_in_log_branch_length = torch.abs(
            log_branch_length - prev_log_branch_length
        )
        if change_in_log_branch_length < optimization_tol:
            break

        prev_log_branch_length = log_branch_length.clone()

    if step_idx == max_optimization_steps - 1:
        print(
            f"Warning: optimization did not converge after {max_optimization_steps} steps; log branch length is {log_branch_length.detach().item()}"
        )
        failed_to_converge = True
    else:
        failed_to_converge = False

    return torch.exp(log_branch_length.detach()).item(), failed_to_converge



import os
import shutil

import fire
import pandas as pd

from netam.pretrained import PRETRAINED_DIR


def clear_model_cache():
    """This function clears the cache of pre-trained models."""
    if os.path.exists(PRETRAINED_DIR):
        print(f"Removing {PRETRAINED_DIR}")
        shutil.rmtree(PRETRAINED_DIR)


def concatenate_csvs(
    input_csvs_str: str,
    output_csv: str,
    is_tsv: bool = False,
    record_path: bool = False,
):
    """This function concatenates multiple CSV or TSV files into one CSV file.

    Args:
        input_csvs: A string of paths to the input CSV or TSV files separated by commas.
        output_csv: Path to the output CSV file.
        is_tsv: A boolean flag that determines whether the input files are TSV.
        record_path: A boolean flag that adds a column recording the path of the input_csv.
    """
    input_csvs = input_csvs_str.split(",")
    dfs = []

    for csv in input_csvs:
        df = pd.read_csv(csv, delimiter="\t" if is_tsv else ",")
        if record_path:
            df["input_file_path"] = csv
        dfs.append(df)

    result_df = pd.concat(dfs, ignore_index=True)

    result_df.to_csv(output_csv, index=False)


def main():
    fire.Fire()



r"""
We are going to get the attention weights using the [MultiHeadAttention](https://pytorch.org/docs/stable/generated/torch.nn.MultiheadAttention.html) module in PyTorch. These weights are

$$
\text{softmax}\left(\frac{Q K^T}{\sqrt{d_k}}\right).
$$

So this tells us that the rows of the attention map correspond to the queries, whereas the columns correspond to the keys.

In our terminology, an attention map is the attention map for a single head. An
"attention maps" object is a collection of attention maps: a tensor where the
first dimension is the number of heads. This assumes all layers have the same
number of heads. An "attention mapss" is a list of attention maps objects, one
for each sequence in the batch. An "attention profile" is some 1-D summary of an
attention map, such as the maximum attention score for each key position.

# Adapted from https://gist.github.com/airalcorn2/50ec06517ce96ecc143503e21fa6cb91
"""

import copy

import torch

from netam.common import aa_idx_tensor_of_str_ambig, aa_mask_tensor_of


def reshape_tensor(tensor, head_count):
    """Reshape the tensor to include the head dimension.

    Assumes batch size is 1 and squeezes it out.
    """
    assert tensor.size(0) == 1, "Batch size should be 1"
    seq_len, embed_dim = tensor.size(1), tensor.size(2)
    head_dim = embed_dim // head_count
    return tensor.view(seq_len, head_count, head_dim).transpose(0, 1)


class SaveAttentionInfo:
    def __init__(self, head_count):
        self.head_count = head_count
        self.attention_maps = []

    def __call__(self, module, module_in, module_out):
        self.attention_maps.append(module_out[1].clone().squeeze(0))

    def clear(self):
        self.attention_maps = []


def patch_attention(m):
    forward_orig = m.forward

    def wrap(*args, **kwargs):
        kwargs["need_weights"] = True
        kwargs["average_attn_weights"] = False

        return forward_orig(*args, **kwargs)

    m.forward = wrap


def attention_mapss_of(model, sequences):
    """Get a list of attention maps (across sequences) as described in the module
    docstring."""
    model = copy.deepcopy(model)
    model.eval()
    layer_count = len(model.encoder.layers)
    # The below assumes all layers have the same number of heads.
    head_count = model.encoder.layers[0].self_attn.num_heads
    save_info = [SaveAttentionInfo(head_count) for _ in range(layer_count)]
    for which_layer, layer in enumerate(model.encoder.layers):
        patch_attention(layer.self_attn)
        layer.self_attn.register_forward_hook(save_info[which_layer])

    for sequence in sequences:
        sequence_idxs = aa_idx_tensor_of_str_ambig(sequence)
        mask = aa_mask_tensor_of(sequence)
        model(sequence_idxs.unsqueeze(0), mask.unsqueeze(0))

    attention_maps = []

    for seq_idx in range(len(sequences)):
        attention_maps.append(
            torch.stack([save.attention_maps[seq_idx] for save in save_info], dim=0)
        )

    return [amap.detach().numpy() for amap in attention_maps]



import pytest
import numpy as np
import torch
from Bio.Seq import Seq
from Bio.Data import CodonTable
from netam.sequences import (
    AA_STR_SORTED,
    CODONS,
    CODON_AA_INDICATOR_MATRIX,
    aa_onehot_tensor_of_str,
    nt_idx_array_of_str,
    nt_subs_indicator_tensor_of,
    translate_sequences,
)


def test_nucleotide_indices_of_codon():
    assert nt_idx_array_of_str("AAA").tolist() == [0, 0, 0]
    assert nt_idx_array_of_str("TAC").tolist() == [3, 0, 1]
    assert nt_idx_array_of_str("GCG").tolist() == [2, 1, 2]


def test_aa_onehot_tensor_of_str():
    aa_str = "QY"

    expected_output = torch.zeros((2, 20))
    expected_output[0][AA_STR_SORTED.index("Q")] = 1
    expected_output[1][AA_STR_SORTED.index("Y")] = 1

    output = aa_onehot_tensor_of_str(aa_str)

    assert output.shape == (2, 20)
    assert torch.equal(output, expected_output)


def test_translate_sequences():
    # sequence without stop codon
    seq_no_stop = ["AGTGGTGGTGGTGGTGGT"]
    assert translate_sequences(seq_no_stop) == [str(Seq(seq_no_stop[0]).translate())]

    # sequence with stop codon
    seq_with_stop = ["TAAGGTGGTGGTGGTAGT"]
    with pytest.raises(ValueError):
        translate_sequences(seq_with_stop)


def test_indicator_matrix():
    reconstructed_codon_table = {}
    indicator_matrix = CODON_AA_INDICATOR_MATRIX.numpy()

    for i, codon in enumerate(CODONS):
        row = indicator_matrix[i]
        if np.any(row):
            amino_acid = AA_STR_SORTED[np.argmax(row)]
            reconstructed_codon_table[codon] = amino_acid

    table = CodonTable.unambiguous_dna_by_id[1]  # 1 is for the standard table

    assert reconstructed_codon_table == table.forward_table


def test_subs_indicator_tensor_of():
    parent = "NAAA"
    child = "CAGA"
    expected_output = torch.tensor([0, 0, 1, 0], dtype=torch.float)
    output = nt_subs_indicator_tensor_of(parent, child)
    assert torch.equal(output, expected_output)



from setuptools import setup, find_packages

# Version is determined by setuptools-scm in pyproject.toml according to git
# tag numbering.
setup(
    name="netam",
    url="https://github.com/matsengrp/netam.git",
    author="Matsen Group",
    author_email="ematsen@gmail.com",
    description="Neural network models for antibody affinity maturation",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    python_requires=">=3.9,<3.12",
    install_requires=[
        "biopython",
        "natsort",
        "optuna",
        "pandas",
        "pyyaml",
        "requests",
        "tensorboardX",
        "torch",
        "tqdm",
        "fire",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.9",
    ],
    entry_points={
        "console_scripts": ["netam=netam.cli:main"],
    },
)



from abc import ABC, abstractmethod
import copy
import multiprocessing as mp
from functools import partial

import torch

# Amazingly, using one thread makes things 50x faster for branch length
# optimization on our server.
torch.set_num_threads(1)

import numpy as np
import pandas as pd

from tqdm import tqdm

from netam.common import (
    MAX_AMBIG_AA_IDX,
    aa_idx_tensor_of_str_ambig,
    stack_heterogeneous,
    codon_mask_tensor_of,
    assert_pcp_valid,
)
import netam.framework as framework
import netam.molevol as molevol
from netam.sequences import (
    aa_subs_indicator_tensor_of,
    translate_sequences,
    apply_aa_mask_to_nt_sequence,
    nt_mutation_frequency,
)


class DXSMDataset(framework.BranchLengthDataset, ABC):
    prefix = "dxsm"

    def __init__(
        self,
        nt_parents: pd.Series,
        nt_children: pd.Series,
        nt_ratess: torch.Tensor,
        nt_cspss: torch.Tensor,
        branch_lengths: torch.Tensor,
        multihit_model=None,
    ):
        self.nt_parents = nt_parents
        self.nt_children = nt_children
        self.nt_ratess = nt_ratess
        self.nt_cspss = nt_cspss
        self.multihit_model = copy.deepcopy(multihit_model)
        if multihit_model is not None:
            # We want these parameters to act like fixed data. This is essential
            # for multithreaded branch length optimization to work.
            self.multihit_model.values.requires_grad_(False)

        assert len(self.nt_parents) == len(self.nt_children)
        pcp_count = len(self.nt_parents)

        aa_parents = translate_sequences(self.nt_parents)
        aa_children = translate_sequences(self.nt_children)
        self.max_aa_seq_len = max(len(seq) for seq in aa_parents)
        # We have sequences of varying length, so we start with all tensors set
        # to the ambiguous amino acid, and then will fill in the actual values
        # below.
        self.aa_parents_idxss = torch.full(
            (pcp_count, self.max_aa_seq_len), MAX_AMBIG_AA_IDX
        )
        self.aa_children_idxss = self.aa_parents_idxss.clone()
        self.aa_subs_indicators = torch.zeros((pcp_count, self.max_aa_seq_len))

        self.masks = torch.ones((pcp_count, self.max_aa_seq_len), dtype=torch.bool)

        for i, (aa_parent, aa_child) in enumerate(zip(aa_parents, aa_children)):
            self.masks[i, :] = codon_mask_tensor_of(
                nt_parents[i], nt_children[i], aa_length=self.max_aa_seq_len
            )
            aa_seq_len = len(aa_parent)
            assert_pcp_valid(
                nt_parents[i], nt_children[i], aa_mask=self.masks[i][:aa_seq_len]
            )

            self.aa_parents_idxss[i, :aa_seq_len] = aa_idx_tensor_of_str_ambig(
                aa_parent
            )
            self.aa_children_idxss[i, :aa_seq_len] = aa_idx_tensor_of_str_ambig(
                aa_child
            )
            self.aa_subs_indicators[i, :aa_seq_len] = aa_subs_indicator_tensor_of(
                aa_parent, aa_child
            )

        assert torch.all(self.masks.sum(dim=1) > 0)
        assert torch.max(self.aa_parents_idxss) <= MAX_AMBIG_AA_IDX

        self._branch_lengths = branch_lengths
        self.update_neutral_probs()

    @classmethod
    def of_seriess(
        cls,
        nt_parents: pd.Series,
        nt_children: pd.Series,
        nt_rates_series: pd.Series,
        nt_csps_series: pd.Series,
        branch_length_multiplier=5.0,
        multihit_model=None,
    ):
        """Alternative constructor that takes the raw data and calculates the initial
        branch lengths.

        The `_series` arguments are series of Tensors which get stacked to
        create the full object.
        """
        initial_branch_lengths = np.array(
            [
                nt_mutation_frequency(parent, child) * branch_length_multiplier
                for parent, child in zip(nt_parents, nt_children)
            ]
        )
        return cls(
            nt_parents.reset_index(drop=True),
            nt_children.reset_index(drop=True),
            stack_heterogeneous(nt_rates_series.reset_index(drop=True)),
            stack_heterogeneous(nt_csps_series.reset_index(drop=True)),
            initial_branch_lengths,
            multihit_model=multihit_model,
        )

    @classmethod
    def of_pcp_df(cls, pcp_df, branch_length_multiplier=5.0, multihit_model=None):
        """Alternative constructor that takes in a pcp_df and calculates the initial
        branch lengths."""
        assert (
            "nt_rates" in pcp_df.columns
        ), "pcp_df must have a neutral nt_rates column"
        return cls.of_seriess(
            pcp_df["parent"],
            pcp_df["child"],
            pcp_df["nt_rates"],
            pcp_df["nt_csps"],
            branch_length_multiplier=branch_length_multiplier,
            multihit_model=multihit_model,
        )

    @classmethod
    def train_val_datasets_of_pcp_df(
        cls, pcp_df, branch_length_multiplier=5.0, multihit_model=None
    ):
        """Perform a train-val split based on the 'in_train' column.

        This is a class method so it works for subclasses.
        """
        train_df = pcp_df[pcp_df["in_train"]].reset_index(drop=True)
        val_df = pcp_df[~pcp_df["in_train"]].reset_index(drop=True)

        val_dataset = cls.of_pcp_df(
            val_df,
            branch_length_multiplier=branch_length_multiplier,
            multihit_model=multihit_model,
        )

        if len(train_df) == 0:
            return None, val_dataset
        # else:
        train_dataset = cls.of_pcp_df(
            train_df,
            branch_length_multiplier=branch_length_multiplier,
            multihit_model=multihit_model,
        )

        return train_dataset, val_dataset

    def clone(self):
        """Make a deep copy of the dataset."""
        new_dataset = self.__class__(
            self.nt_parents,
            self.nt_children,
            self.nt_ratess.copy(),
            self.nt_cspss.copy(),
            self._branch_lengths.copy(),
            multihit_model=self.multihit_model,
        )
        return new_dataset

    def subset_via_indices(self, indices):
        """Create a new dataset with a subset of the data, as per `indices`.

        Whether the new dataset is a deep copy or a shallow copy using slices
        depends on `indices`: if `indices` is an iterable of integers, then we
        make a deep copy, otherwise we use slices to make a shallow copy.
        """
        new_dataset = self.__class__(
            self.nt_parents[indices].reset_index(drop=True),
            self.nt_children[indices].reset_index(drop=True),
            self.nt_ratess[indices],
            self.nt_cspss[indices],
            self._branch_lengths[indices],
            multihit_model=self.multihit_model,
        )
        return new_dataset

    def split(self, into_count: int):
        """Split self into a list of into_count subsets."""
        dataset_size = len(self)
        indices = list(range(dataset_size))
        split_indices = np.array_split(indices, into_count)
        subsets = [self.subset_via_indices(split_indices[i]) for i in range(into_count)]
        return subsets

    @property
    def branch_lengths(self):
        return self._branch_lengths

    @branch_lengths.setter
    def branch_lengths(self, new_branch_lengths):
        assert len(new_branch_lengths) == len(self._branch_lengths), (
            f"Expected {len(self._branch_lengths)} branch lengths, "
            f"got {len(new_branch_lengths)}"
        )
        assert torch.all(torch.isfinite(new_branch_lengths) & (new_branch_lengths > 0))
        self._branch_lengths = new_branch_lengths
        self.update_neutral_probs()

    @abstractmethod
    def update_neutral_probs(self):
        pass


class DXSMBurrito(framework.Burrito, ABC):
    prefix = "dxsm"

    def _find_optimal_branch_length(
        self,
        parent,
        child,
        nt_rates,
        nt_csps,
        aa_mask,
        starting_branch_length,
        multihit_model,
        **optimization_kwargs,
    ):
        sel_matrix = self.build_selection_matrix_from_parent(parent)
        trimmed_aa_mask = aa_mask[: len(sel_matrix)]
        log_pcp_probability = molevol.mutsel_log_pcp_probability_of(
            sel_matrix[trimmed_aa_mask],
            apply_aa_mask_to_nt_sequence(parent, trimmed_aa_mask),
            apply_aa_mask_to_nt_sequence(child, trimmed_aa_mask),
            nt_rates[trimmed_aa_mask.repeat_interleave(3)],
            nt_csps[trimmed_aa_mask.repeat_interleave(3)],
            multihit_model,
        )
        if isinstance(starting_branch_length, torch.Tensor):
            starting_branch_length = starting_branch_length.detach().item()
        return molevol.optimize_branch_length(
            log_pcp_probability, starting_branch_length, **optimization_kwargs
        )

    def serial_find_optimal_branch_lengths(self, dataset, **optimization_kwargs):
        optimal_lengths = []
        failed_count = 0

        for parent, child, nt_rates, nt_csps, aa_mask, starting_length in tqdm(
            zip(
                dataset.nt_parents,
                dataset.nt_children,
                dataset.nt_ratess,
                dataset.nt_cspss,
                dataset.masks,
                dataset.branch_lengths,
            ),
            total=len(dataset.nt_parents),
            desc="Finding optimal branch lengths",
        ):
            branch_length, failed_to_converge = self._find_optimal_branch_length(
                parent,
                child,
                nt_rates[: len(parent)],
                nt_csps[: len(parent), :],
                aa_mask,
                starting_length,
                dataset.multihit_model,
                **optimization_kwargs,
            )

            optimal_lengths.append(branch_length)
            failed_count += failed_to_converge

        if failed_count > 0:
            print(
                f"Branch length optimization failed to converge for {failed_count} of {len(dataset)} sequences."
            )

        return torch.tensor(optimal_lengths)

    def find_optimal_branch_lengths(self, dataset, **optimization_kwargs):
        worker_count = min(mp.cpu_count() // 2, 10)
        # # The following can be used when one wants a better traceback.
        # burrito = self.__class__(None, dataset, copy.deepcopy(self.model))
        # return burrito.serial_find_optimal_branch_lengths(dataset, **optimization_kwargs)
        our_optimize_branch_length = partial(
            worker_optimize_branch_length,
            self.__class__,
        )
        with mp.Pool(worker_count) as pool:
            splits = dataset.split(worker_count)
            results = pool.starmap(
                our_optimize_branch_length,
                [(self.model, split, optimization_kwargs) for split in splits],
            )
        return torch.cat(results)

    def load_branch_lengths(self, in_csv_prefix):
        if self.train_dataset is not None:
            self.train_dataset.load_branch_lengths(
                in_csv_prefix + ".train_branch_lengths.csv"
            )
        self.val_dataset.load_branch_lengths(in_csv_prefix + ".val_branch_lengths.csv")

    def to_crepe(self):
        training_hyperparameters = {
            key: self.__dict__[key]
            for key in [
                "optimizer_name",
                "batch_size",
                "learning_rate",
                "min_learning_rate",
                "weight_decay",
            ]
        }
        encoder = framework.PlaceholderEncoder()
        return framework.Crepe(encoder, self.model, training_hyperparameters)

    @abstractmethod
    def loss_of_batch(self, batch):
        pass


def worker_optimize_branch_length(burrito_class, model, dataset, optimization_kwargs):
    """The worker used for parallel branch length optimization."""
    burrito = burrito_class(None, dataset, copy.deepcopy(model))
    return burrito.serial_find_optimal_branch_lengths(dataset, **optimization_kwargs)



import os

import netam.multihit as multihit
import netam.framework as framework
import netam.hit_class as hit_class
from netam.molevol import (
    codon_probs_of_parent_scaled_nt_rates_and_csps,
    reshape_for_codons,
)
from netam import pretrained
from netam.sequences import nt_idx_tensor_of_str
import pytest
import pandas as pd
import torch

burrito_params = {
    "batch_size": 1024,
    "learning_rate": 0.1,
    "min_learning_rate": 1e-4,
}


# These happen to be the same as some examples in test_models.py but that's fine.
# If it was important that they were shared, we should put them in a conftest.py.
ex_scaled_rates = torch.tensor([0.01, 0.001, 0.005])
ex_csps = torch.tensor(
    [[0.0, 0.3, 0.5, 0.2], [0.4, 0.0, 0.1, 0.5], [0.2, 0.3, 0.0, 0.5]]
)
# This is an example, and the correct output for test_codon_probs_of_parent_scaled_nt_rates_and_csps
ex_codon_probs = torch.tensor(
    [
        [
            [
                [3.9484e-07, 5.9226e-07, 3.9385e-04, 9.8710e-07],
                [9.8660e-04, 1.4799e-03, 9.8413e-01, 2.4665e-03],
                [9.8710e-08, 1.4806e-07, 9.8463e-05, 2.4677e-07],
                [4.9355e-07, 7.4032e-07, 4.9231e-04, 1.2339e-06],
            ],
            [
                [1.1905e-09, 1.7857e-09, 1.1875e-06, 2.9762e-09],
                [2.9746e-06, 4.4619e-06, 2.9672e-03, 7.4366e-06],
                [2.9762e-10, 4.4642e-10, 2.9687e-07, 7.4404e-10],
                [1.4881e-09, 2.2321e-09, 1.4844e-06, 3.7202e-09],
            ],
            [
                [1.9841e-09, 2.9762e-09, 1.9791e-06, 4.9602e-09],
                [4.9577e-06, 7.4366e-06, 4.9453e-03, 1.2394e-05],
                [4.9602e-10, 7.4404e-10, 4.9478e-07, 1.2401e-09],
                [2.4801e-09, 3.7202e-09, 2.4739e-06, 6.2003e-09],
            ],
            [
                [7.9364e-10, 1.1905e-09, 7.9165e-07, 1.9841e-09],
                [1.9831e-06, 2.9746e-06, 1.9781e-03, 4.9577e-06],
                [1.9841e-10, 2.9762e-10, 1.9791e-07, 4.9602e-10],
                [9.9205e-10, 1.4881e-09, 9.8957e-07, 2.4801e-09],
            ],
        ]
    ]
)

ex_parent_codon_idxs = nt_idx_tensor_of_str("ACG")


@pytest.fixture
def mini_multihit_train_val_datasets():
    df = pd.read_csv("data/wyatt-10x-1p5m_pcp_2023-11-30_NI.first100.csv.gz")
    crepe = pretrained.load("ThriftyHumV0.2-45")
    df = multihit.prepare_pcp_df(df, crepe, 500)
    return multihit.train_test_datasets_of_pcp_df(df)


@pytest.fixture
def hitclass_burrito(mini_multihit_train_val_datasets):
    train_data, val_data = mini_multihit_train_val_datasets
    return multihit.MultihitBurrito(
        train_data, val_data, multihit.HitClassModel(), **burrito_params
    )


def test_train(hitclass_burrito):
    before_values = hitclass_burrito.model.values.clone()
    hitclass_burrito.joint_train(epochs=2)
    assert not torch.allclose(hitclass_burrito.model.values, before_values)


def test_serialize(hitclass_burrito):
    os.makedirs("_ignore", exist_ok=True)
    hitclass_burrito.save_crepe("_ignore/test_multihit_crepe")
    new_crepe = framework.load_crepe("_ignore/test_multihit_crepe")
    assert torch.allclose(new_crepe.model.values, hitclass_burrito.model.values)


def test_codon_probs_of_parent_scaled_nt_rates_and_csps():
    computed_tensor = codon_probs_of_parent_scaled_nt_rates_and_csps(
        ex_parent_codon_idxs, ex_scaled_rates, ex_csps
    )
    correct_tensor = ex_codon_probs
    assert torch.allclose(correct_tensor, computed_tensor)
    assert torch.allclose(
        computed_tensor.sum(dim=(1, 2, 3)), torch.ones(computed_tensor.shape[0])
    )


def test_multihit_correction():
    hit_class_factors = torch.tensor([-0.1, 1, 2.3])
    # We'll verify that aggregating by hit class then adjusting is the same as adjusting then aggregating by hit class.
    codon_idxs = reshape_for_codons(ex_parent_codon_idxs)
    adjusted_codon_probs = hit_class.apply_multihit_correction(
        codon_idxs, ex_codon_probs, hit_class_factors
    )
    aggregate_last = hit_class.hit_class_probs_tensor(codon_idxs, adjusted_codon_probs)

    uncorrected_hc_log_probs = hit_class.hit_class_probs_tensor(
        codon_idxs, ex_codon_probs
    ).log()

    corrections = torch.cat([torch.tensor([0.0]), hit_class_factors])
    # we'll use the corrections to adjust the uncorrected hit class probs
    corrections = corrections[
        torch.arange(4).unsqueeze(0).tile((uncorrected_hc_log_probs.shape[0], 1))
    ]
    uncorrected_hc_log_probs += corrections
    aggregate_first = torch.softmax(uncorrected_hc_log_probs, dim=1)
    assert torch.allclose(aggregate_first, aggregate_last)


def test_hit_class_tensor():
    # verify that the opaque way of computing the hit class tensor is the same
    # as the transparent way.
    def compute_hit_class(codon1, codon2):
        return sum(c1 != c2 for c1, c2 in zip(codon1, codon2))

    true_hit_class_tensor = torch.zeros(4, 4, 4, 4, 4, 4, dtype=torch.int)

    # Populate the tensor
    for i1 in range(4):
        for j1 in range(4):
            for k1 in range(4):
                codon1 = (i1, j1, k1)
                for i2 in range(4):
                    for j2 in range(4):
                        for k2 in range(4):
                            codon2 = (i2, j2, k2)
                            true_hit_class_tensor[i1, j1, k1, i2, j2, k2] = (
                                compute_hit_class(codon1, codon2)
                            )
    assert torch.allclose(hit_class.hit_class_tensor, true_hit_class_tensor)



import pytest

from netam.common import force_spawn
from netam.models import TransformerBinarySelectionModelWiggleAct
from netam.dasm import (
    DASMBurrito,
    DASMDataset,
)
from netam.dnsm import DNSMBurrito, DNSMDataset
from netam.framework import (
    load_pcp_df,
    add_shm_model_outputs_to_pcp_df,
)
from netam import pretrained
import random


# Function to randomly insert 'N' in sequences
def randomize_with_ns(parent_seq, child_seq, avoid_masked_equality=True):
    old_parent = parent_seq
    old_child = child_seq
    seq_length = len(parent_seq)
    try:
        first_mut = next(
            (idx, p, c)
            for idx, (p, c) in enumerate(zip(parent_seq, child_seq))
            if p != c
        )
    except:
        return parent_seq, child_seq

    # Decide which type of modification to apply
    modification_type = random.choice(["same_site", "different_site", "chunk", "none"])

    if modification_type == "same_site":
        # Select random positions to place 'N' in both parent and child at the same positions
        num_ns = random.randint(
            1, seq_length // 3
        )  # Random number of Ns, max third the sequence length
        positions = random.sample(range(seq_length), num_ns)
        parent_seq = "".join(
            ["N" if i in positions else base for i, base in enumerate(parent_seq)]
        )
        child_seq = "".join(
            ["N" if i in positions else base for i, base in enumerate(child_seq)]
        )

    elif modification_type == "different_site":
        # Insert 'N's at random positions in parent and child, but not the same positions
        num_ns_parent = random.randint(1, seq_length // 3)
        num_ns_child = random.randint(1, seq_length // 3)
        positions_parent = random.sample(range(seq_length), num_ns_parent)
        positions_child = random.sample(range(seq_length), num_ns_child)

        parent_seq = "".join(
            [
                "N" if i in positions_parent else base
                for i, base in enumerate(parent_seq)
            ]
        )
        child_seq = "".join(
            ["N" if i in positions_child else base for i, base in enumerate(child_seq)]
        )

    elif modification_type == "chunk":
        # Replace a chunk of bases with 'N's in both parent and child
        chunk_size = random.randint(1, seq_length // 3)
        start_pos = random.randint(0, seq_length - chunk_size)
        parent_seq = (
            parent_seq[:start_pos]
            + "N" * chunk_size
            + parent_seq[start_pos + chunk_size :]
        )
        child_seq = (
            child_seq[:start_pos]
            + "N" * chunk_size
            + child_seq[start_pos + chunk_size :]
        )

    if parent_seq == child_seq:
        # If sequences are the same, put one mutated site back in:
        idx, p, c = first_mut
        parent_seq = parent_seq[:idx] + p + parent_seq[idx + 1 :]
        child_seq = child_seq[:idx] + c + child_seq[idx + 1 :]
    if avoid_masked_equality:
        codon_pairs = [
            (parent_seq[i * 3 : (i + 1) * 3], child_seq[i * 3 : (i + 1) * 3])
            for i in range(seq_length // 3)
        ]
        if all(
            p == c
            for p, c in filter(
                lambda pair: "N" not in pair[0] and "N" not in pair[1], codon_pairs
            )
        ):
            # put original codon containing a mutation back in.
            idx, p, c = first_mut
            codon_start = (idx // 3) * 3
            codon_end = codon_start + 3
            parent_seq = (
                parent_seq[:codon_start]
                + old_parent[codon_start:codon_end]
                + parent_seq[codon_end:]
            )
            child_seq = (
                child_seq[:codon_start]
                + old_child[codon_start:codon_end]
                + child_seq[codon_end:]
            )

    assert len(parent_seq) == len(child_seq)
    assert len(parent_seq) == seq_length

    return parent_seq, child_seq


@pytest.fixture
def ambig_pcp_df():
    random.seed(1)
    df = load_pcp_df(
        "data/wyatt-10x-1p5m_pcp_2023-11-30_NI.first100.csv.gz",
    )
    # Apply the random N adding function to each row
    df[["parent", "child"]] = df.apply(
        lambda row: randomize_with_ns(row["parent"], row["child"]),
        axis=1,
        result_type="expand",
    )

    df = add_shm_model_outputs_to_pcp_df(
        df,
        pretrained.load("ThriftyHumV0.2-45"),
    )
    return df


@pytest.fixture
def dnsm_model():
    return TransformerBinarySelectionModelWiggleAct(
        nhead=2, d_model_per_head=4, dim_feedforward=256, layer_count=2
    )


def test_dnsm_burrito(ambig_pcp_df, dnsm_model):
    """Fixture that returns the DNSM Burrito object."""
    force_spawn()
    ambig_pcp_df["in_train"] = True
    ambig_pcp_df.loc[ambig_pcp_df.index[-15:], "in_train"] = False
    train_dataset, val_dataset = DNSMDataset.train_val_datasets_of_pcp_df(ambig_pcp_df)

    burrito = DNSMBurrito(
        train_dataset,
        val_dataset,
        dnsm_model,
        batch_size=32,
        learning_rate=0.001,
        min_learning_rate=0.0001,
    )
    burrito.joint_train(epochs=1, cycle_count=2, training_method="full")


@pytest.fixture
def dasm_model():
    return TransformerBinarySelectionModelWiggleAct(
        nhead=2,
        d_model_per_head=4,
        dim_feedforward=256,
        layer_count=2,
        output_dim=20,
    )


def test_dasm_burrito(ambig_pcp_df, dasm_model):
    force_spawn()
    """Fixture that returns the DNSM Burrito object."""
    ambig_pcp_df["in_train"] = True
    ambig_pcp_df.loc[ambig_pcp_df.index[-15:], "in_train"] = False
    train_dataset, val_dataset = DASMDataset.train_val_datasets_of_pcp_df(ambig_pcp_df)

    burrito = DASMBurrito(
        train_dataset,
        val_dataset,
        dasm_model,
        batch_size=32,
        learning_rate=0.001,
        min_learning_rate=0.0001,
    )
    burrito.joint_train(
        epochs=1, cycle_count=2, training_method="full", optimize_bl_first_cycle=False
    )



import os

import torch
import pytest

from netam.framework import (
    crepe_exists,
    load_crepe,
)
from netam.common import aa_idx_tensor_of_str_ambig, MAX_AMBIG_AA_IDX, force_spawn
from netam.models import TransformerBinarySelectionModelWiggleAct
from netam.dnsm import DNSMBurrito, DNSMDataset


def test_aa_idx_tensor_of_str_ambig():
    input_seq = "ACX"
    expected_output = torch.tensor([0, 1, MAX_AMBIG_AA_IDX], dtype=torch.int)
    output = aa_idx_tensor_of_str_ambig(input_seq)
    assert torch.equal(output, expected_output)


@pytest.fixture(scope="module")
def dnsm_burrito(pcp_df):
    """Fixture that returns the DNSM Burrito object."""
    force_spawn()
    pcp_df["in_train"] = True
    pcp_df.loc[pcp_df.index[-15:], "in_train"] = False
    train_dataset, val_dataset = DNSMDataset.train_val_datasets_of_pcp_df(pcp_df)

    model = TransformerBinarySelectionModelWiggleAct(
        nhead=2, d_model_per_head=4, dim_feedforward=256, layer_count=2
    )

    burrito = DNSMBurrito(
        train_dataset,
        val_dataset,
        model,
        batch_size=32,
        learning_rate=0.001,
        min_learning_rate=0.0001,
    )
    burrito.joint_train(epochs=1, cycle_count=2, training_method="full")
    return burrito


def test_parallel_branch_length_optimization(dnsm_burrito):
    force_spawn()
    dataset = dnsm_burrito.val_dataset
    parallel_branch_lengths = dnsm_burrito.find_optimal_branch_lengths(dataset)
    branch_lengths = dnsm_burrito.serial_find_optimal_branch_lengths(dataset)
    assert torch.allclose(branch_lengths, parallel_branch_lengths)


def test_crepe_roundtrip(dnsm_burrito):
    os.makedirs("_ignore", exist_ok=True)
    crepe_path = "_ignore/dnsm"
    dnsm_burrito.save_crepe(crepe_path)
    assert crepe_exists(crepe_path)
    crepe = load_crepe(crepe_path)
    model = crepe.model
    assert isinstance(model, TransformerBinarySelectionModelWiggleAct)
    assert dnsm_burrito.model.hyperparameters == model.hyperparameters
    model.to(dnsm_burrito.device)
    for t1, t2 in zip(
        dnsm_burrito.model.state_dict().values(), model.state_dict().values()
    ):
        assert torch.equal(t1, t2)
